from datetime import date
from typing import Optional
from unittest import IsolatedAsyncioTestCase, TestCase
from unittest.mock import call, sentinel

from asyncstdlib import iter as aiter, tuple as atuple
from fred_api.registry.contact.contact_history_types_pb2 import (
    ContactHistoryRecord,
    ContactHistoryReply,
    ContactHistoryRequest,
)
from fred_api.registry.contact.contact_info_types_pb2 import (
    Birthdate,
    ContactAdditionalIdentifier as AdditionalIdentifierMessage,
    ContactIdReply,
    ContactIdRequest,
    ContactInfoReply,
    ContactInfoRequest,
    WarningLetter as WarningLetterMessage,
)
from fred_api.registry.contact.contact_state_flags_types_pb2 import ContactStateFlagsReply
from fred_api.registry.contact.contact_state_history_types_pb2 import (
    ContactStateHistory,
    ContactStateHistoryReply,
    ContactStateHistoryRequest,
)
from fred_api.registry.contact.contact_state_types_pb2 import ContactStateReply, ContactStateRequest
from fred_api.registry.contact.list_contacts_types_pb2 import ListContactsReply
from fred_api.registry.contact.update_contact_state_types_pb2 import UpdateContactStateReply, UpdateContactStateRequest
from fred_types import ContactHistoryId, ContactId, ContactRef, LogEntryId
from frgal.utils import AsyncTestClientMixin, make_awaitable
from google.protobuf.empty_pb2 import Empty

from regal.contact import (
    AdditionalIdentifier,
    Contact,
    ContactClient,
    ContactDecoder,
    ContactIdentifierType,
    WarningLetterPreference,
)
from regal.exceptions import ContactDoesNotExist, NotCurrentVersion
from regal.tests.common import ObjectClientTestMixin


class AdditionalIdentifierTest(TestCase):
    def test_birthdate(self):
        data = (
            # ident, birthdate
            (AdditionalIdentifier(type=ContactIdentifierType.BIRTHDATE, value="1988-02-15"), date(1988, 2, 15)),
            (AdditionalIdentifier(type=ContactIdentifierType.BIRTHDATE, value=""), None),
            (AdditionalIdentifier(type=ContactIdentifierType.PASSPORT_NUMBER, value="1988-02-15"), None),
        )
        for ident, result in data:
            with self.subTest(ident=ident):
                self.assertEqual(ident.birthdate, result)

    def test_birthdate_invalid(self):
        ident = AdditionalIdentifier(type=ContactIdentifierType.BIRTHDATE, value="JUNK")
        with self.assertRaisesRegex(ValueError, "'JUNK' is not a valid date."):
            ident.birthdate


class ContactDecoderTest(IsolatedAsyncioTestCase):
    """Test ContactDecoder."""

    def setUp(self):
        self.decoder = ContactDecoder()

    def test_additional_identifier(self):
        self.assertEqual(
            self.decoder.decode(AdditionalIdentifierMessage(birthdate=Birthdate(value="2000-01-01"))),
            AdditionalIdentifier(type=ContactIdentifierType.BIRTHDATE, value="2000-01-01"),
        )
        self.assertIsNone(self.decoder.decode(AdditionalIdentifierMessage()))

    def test_decode_warning_letter(self):
        self.assertEqual(self.decoder.decode(WarningLetterMessage(preference=0)), WarningLetterPreference.TO_SEND)
        self.assertEqual(self.decoder.decode(WarningLetterMessage(preference=1)), WarningLetterPreference.NOT_TO_SEND)
        self.assertEqual(self.decoder.decode(WarningLetterMessage(preference=2)), WarningLetterPreference.NOT_SPECIFIED)


class TestContactClient(AsyncTestClientMixin, ContactClient):
    """Test ContactClient."""


class ContactClientTest(ObjectClientTestMixin[TestContactClient], IsolatedAsyncioTestCase):
    """Test ContactClient's methods."""

    grpc_namespace = "/Fred.Registry.Api.Contact.Contact"
    get_state_request = ContactStateRequest
    get_state_reply = ContactStateReply
    get_state_method = "get_contact_state"
    get_history_request = ContactHistoryRequest
    get_history_reply = ContactHistoryReply
    get_history_method = "get_contact_history"
    get_history_record = ContactHistoryRecord
    get_state_history_request = ContactStateHistoryRequest
    get_state_history_reply = ContactStateHistoryReply
    get_state_history_method = "get_contact_state_history"
    get_state_history_record = ContactStateHistory.Record
    get_state_flags_method = "get_contact_state_flags"
    get_state_flags_reply_cls = ContactStateFlagsReply

    def setUp(self):
        self.client = TestContactClient(netloc=sentinel.netloc)

    async def test_get_contact_id(self):
        contact_id = ContactId("kryten")
        reply = ContactIdReply()
        reply.data.contact_id.uuid.value = str(contact_id)
        self.client.mock.return_value = make_awaitable(reply)

        response = await self.client.get_contact_id("HANDLE")
        self.assertEqual(response, contact_id)
        self.assertIsInstance(response, ContactId)

        request = ContactIdRequest()
        request.contact_handle.value = "HANDLE"
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.Registry.Api.Contact.Contact/get_contact_id",
            timeout=None,
        )

    async def _test_get_contact_info(self, contact_history_id: Optional[ContactHistoryId] = None) -> None:
        reply = ContactInfoReply()
        reply.data.contact_id.uuid.value = "kryten"
        if contact_history_id:
            reply.data.contact_history_id.uuid.value = contact_history_id
        reply.data.contact_handle.value = "HANDLE"
        reply.data.warning_letter.preference = WarningLetterPreference.NOT_TO_SEND.value
        reply.data.additional_identifier.birthdate.value = "2000-01-01"
        self.client.mock.return_value = make_awaitable(reply)

        response = await self.client.get_contact_info(ContactId("kryten"), history_id=contact_history_id)

        self.assertEqual(
            response,
            Contact(
                contact_id=ContactId("kryten"),
                contact_history_id=contact_history_id,
                contact_handle="HANDLE",
                additional_identifier=AdditionalIdentifier(type=ContactIdentifierType.BIRTHDATE, value="2000-01-01"),
                warning_letter=WarningLetterPreference.NOT_TO_SEND,
            ),
        )

        request = ContactInfoRequest()
        request.contact_id.uuid.value = "kryten"
        if contact_history_id:
            request.contact_history_id.uuid.value = contact_history_id
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.Registry.Api.Contact.Contact/get_contact_info",
            timeout=None,
        )

    async def test_get_contact_info(self):
        await self._test_get_contact_info()

    async def test_get_contact_info_with_history_id(self):
        await self._test_get_contact_info(contact_history_id=ContactHistoryId("nova-5"))

    async def test_list_empty_stream(self):
        self.client.mock.return_value = aiter(())

        contacts = await atuple(self.client.list())

        self.assertEqual(contacts, ())
        self.assertEqual(
            self.client.mock.mock_calls,
            [call(Empty(), method="/Fred.Registry.Api.Contact.Contact/list_contacts", timeout=None)],
        )

    async def test_list_empty_response(self):
        reply = ListContactsReply()
        reply.data.contacts.extend([])
        self.client.mock.return_value = aiter([reply])

        contacts = await atuple(self.client.list())

        self.assertEqual(contacts, ())
        self.assertEqual(
            self.client.mock.mock_calls,
            [call(Empty(), method="/Fred.Registry.Api.Contact.Contact/list_contacts", timeout=None)],
        )

    async def test_list(self):
        reply = ListContactsReply()
        contact_ref = reply.data.contacts.add()
        contact_ref.id.uuid.value = "42"
        contact_ref.handle.value = "KRYTEN"
        self.client.mock.return_value = aiter([reply])

        contacts = await atuple(self.client.list())

        self.assertEqual(contacts, (ContactRef(id=ContactId("42"), handle="KRYTEN"),))
        self.assertEqual(
            self.client.mock.mock_calls,
            [call(Empty(), method="/Fred.Registry.Api.Contact.Contact/list_contacts", timeout=None)],
        )

    async def test_list_flatten(self):
        # Test responses are properly flattened.
        reply_1 = ListContactsReply()
        contact_ref_1 = reply_1.data.contacts.add()
        contact_ref_1.id.uuid.value = "42"
        contact_ref_1.handle.value = "KRYTEN"
        contact_ref_2 = reply_1.data.contacts.add()
        contact_ref_2.id.uuid.value = "43"
        contact_ref_2.handle.value = "LISTER"
        reply_2 = ListContactsReply()
        contact_ref_3 = reply_2.data.contacts.add()
        contact_ref_3.id.uuid.value = "44"
        contact_ref_3.handle.value = "RIMMER"
        contact_ref_4 = reply_2.data.contacts.add()
        contact_ref_4.id.uuid.value = "45"
        contact_ref_4.handle.value = "CAT"
        self.client.mock.return_value = aiter([reply_1, reply_2])

        contacts = await atuple(self.client.list())

        result = (
            ContactRef(id=ContactId("42"), handle="KRYTEN"),
            ContactRef(id=ContactId("43"), handle="LISTER"),
            ContactRef(id=ContactId("44"), handle="RIMMER"),
            ContactRef(id=ContactId("45"), handle="CAT"),
        )
        self.assertEqual(contacts, result)
        self.assertEqual(
            self.client.mock.mock_calls,
            [call(Empty(), method="/Fred.Registry.Api.Contact.Contact/list_contacts", timeout=None)],
        )

    async def test_update_contact_state(self):
        flags = {"state 1": True, "state 2": True, "State 3": False}
        reply = UpdateContactStateReply()
        self.client.mock.return_value = make_awaitable(reply)

        await self.client.update_contact_state(ContactId("Id-1"), ContactHistoryId("Hid-1"), flags, LogEntryId("Log-1"))

        request = UpdateContactStateRequest()
        request.contact_id.uuid.value = "Id-1"
        request.contact_history_id.uuid.value = "Hid-1"
        request.log_entry_id.value = "Log-1"
        for key, value in flags.items():
            request.set_flags[key] = value
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.Registry.Api.Contact.Contact/update_contact_state",
            timeout=None,
        )

    async def test_update_contact_state_not_found(self):
        reply = UpdateContactStateReply()
        reply.exception.contact_does_not_exist.CopyFrom(UpdateContactStateReply.Exception.ContactDoesNotExist())
        self.client.mock.return_value = make_awaitable(reply)

        with self.assertRaises(ContactDoesNotExist):
            await self.client.update_contact_state(ContactId("Id-1"), ContactHistoryId("Hid-1"), {})

    async def test_update_contact_state_not_current(self):
        reply = UpdateContactStateReply()
        reply.exception.not_current_version.CopyFrom(UpdateContactStateReply.Exception.NotCurrentVersion())
        self.client.mock.return_value = make_awaitable(reply)

        with self.assertRaises(NotCurrentVersion):
            await self.client.update_contact_state(ContactId("Id-1"), ContactHistoryId("Hid-1"), {})
