from ipaddress import IPv4Address
from typing import Optional
from unittest import IsolatedAsyncioTestCase
from unittest.mock import call, sentinel

from asyncstdlib import iter as aiter, tuple as atuple
from fred_api.registry.nsset.list_nssets_types_pb2 import ListNssetsReply
from fred_api.registry.nsset.nsset_history_types_pb2 import NssetHistoryRecord, NssetHistoryReply, NssetHistoryRequest
from fred_api.registry.nsset.nsset_info_types_pb2 import (
    CheckDnsHostReply,
    CheckDnsHostRequest,
    DnsHost as DnsHostMessage,
    NssetIdReply,
    NssetIdRequest,
    NssetInfoReply,
    NssetInfoRequest,
)
from fred_api.registry.nsset.nsset_state_flags_types_pb2 import NssetStateFlagsReply
from fred_api.registry.nsset.nsset_state_history_types_pb2 import (
    NssetStateHistory,
    NssetStateHistoryReply,
    NssetStateHistoryRequest,
)
from fred_api.registry.nsset.nsset_state_types_pb2 import NssetStateReply, NssetStateRequest
from fred_types import NssetRef
from frgal.utils import AsyncTestClientMixin, make_awaitable
from google.protobuf.empty_pb2 import Empty

from regal.nsset import DnsHost, Nsset, NssetClient, NssetHistoryId, NssetId
from regal.tests.common import ObjectClientTestMixin


class TestNssetClient(AsyncTestClientMixin, NssetClient):
    """Test NssetClient."""


class NssetClientTest(ObjectClientTestMixin[TestNssetClient], IsolatedAsyncioTestCase):
    """Test NssetClient's methods."""

    grpc_namespace = "/Fred.Registry.Api.Nsset.Nsset"
    get_state_request = NssetStateRequest
    get_state_reply = NssetStateReply
    get_state_method = "get_nsset_state"
    get_history_request = NssetHistoryRequest
    get_history_reply = NssetHistoryReply
    get_history_method = "get_nsset_history"
    get_history_record = NssetHistoryRecord
    get_state_history_request = NssetStateHistoryRequest
    get_state_history_reply = NssetStateHistoryReply
    get_state_history_method = "get_nsset_state_history"
    get_state_history_record = NssetStateHistory.Record
    get_state_flags_method = "get_nsset_state_flags"
    get_state_flags_reply_cls = NssetStateFlagsReply

    def setUp(self):
        self.client = TestNssetClient(netloc=sentinel.netloc)

    async def test_list_empty_stream(self):
        self.client.mock.return_value = aiter(())

        response = await atuple(self.client.list())

        self.assertEqual(response, ())
        self.assertEqual(
            self.client.mock.mock_calls,
            [call(Empty(), method="/Fred.Registry.Api.Nsset.Nsset/list_nssets", timeout=None)],
        )

    async def test_list_empty_response(self):
        reply = ListNssetsReply()
        reply.data.nssets.extend([])
        self.client.mock.return_value = aiter([reply])

        response = await atuple(self.client.list())

        self.assertEqual(response, ())
        self.assertEqual(
            self.client.mock.mock_calls,
            [call(Empty(), method="/Fred.Registry.Api.Nsset.Nsset/list_nssets", timeout=None)],
        )

    async def test_list(self):
        reply = ListNssetsReply()
        nsset_ref = reply.data.nssets.add()
        nsset_ref.id.uuid.value = "42"
        nsset_ref.handle.value = "GAZPACHO"
        self.client.mock.return_value = aiter([reply])

        response = await atuple(self.client.list())

        self.assertEqual(response, (NssetRef(id=NssetId("42"), handle="GAZPACHO"),))
        self.assertEqual(
            self.client.mock.mock_calls,
            [call(Empty(), method="/Fred.Registry.Api.Nsset.Nsset/list_nssets", timeout=None)],
        )

    async def test_list_flatten(self):
        # Test responses are properly flattened.
        reply_1 = ListNssetsReply()
        nsset_ref_1 = reply_1.data.nssets.add()
        nsset_ref_1.id.uuid.value = "42"
        nsset_ref_1.handle.value = "GAZPACHO.1"
        nsset_ref_2 = reply_1.data.nssets.add()
        nsset_ref_2.id.uuid.value = "43"
        nsset_ref_2.handle.value = "GAZPACHO.2"
        reply_2 = ListNssetsReply()
        nsset_ref_3 = reply_2.data.nssets.add()
        nsset_ref_3.id.uuid.value = "44"
        nsset_ref_3.handle.value = "GAZPACHO.3"
        nsset_ref_4 = reply_2.data.nssets.add()
        nsset_ref_4.id.uuid.value = "45"
        nsset_ref_4.handle.value = "GAZPACHO.4"
        self.client.mock.return_value = aiter([reply_1, reply_2])

        response = await atuple(self.client.list())

        expected = (
            NssetRef(id=NssetId("42"), handle="GAZPACHO.1"),
            NssetRef(id=NssetId("43"), handle="GAZPACHO.2"),
            NssetRef(id=NssetId("44"), handle="GAZPACHO.3"),
            NssetRef(id=NssetId("45"), handle="GAZPACHO.4"),
        )
        self.assertEqual(response, expected)
        self.assertEqual(
            self.client.mock.mock_calls,
            [call(Empty(), method="/Fred.Registry.Api.Nsset.Nsset/list_nssets", timeout=None)],
        )

    async def test_get_nsset_id(self):
        nsset_id = NssetId("kryten")
        reply = NssetIdReply()
        reply.data.nsset_id.uuid.value = str(nsset_id)
        self.client.mock.return_value = make_awaitable(reply)

        response = await self.client.get_nsset_id("HANDLE")
        self.assertEqual(response, nsset_id)
        self.assertIsInstance(response, NssetId)

        request = NssetIdRequest()
        request.nsset_handle.value = "HANDLE"
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.Registry.Api.Nsset.Nsset/get_nsset_id",
            timeout=None,
        )

    async def _test_get_nsset_info(self, nsset_history_id: Optional[NssetHistoryId] = None) -> None:
        dns_host = DnsHostMessage()
        dns_host.fqdn = "dns.example.com"
        dns_host.ip_addresses[:] = ["127.0.0.1"]
        reply = NssetInfoReply()
        reply.data.nsset_id.uuid.value = "kryten"
        if nsset_history_id:
            reply.data.nsset_history_id.uuid.value = nsset_history_id
        reply.data.nsset_handle.value = "HANDLE"
        reply.data.dns_hosts.extend([dns_host])
        self.client.mock.return_value = make_awaitable(reply)

        response = await self.client.get_nsset_info(NssetId("kryten"), history_id=nsset_history_id)

        self.assertEqual(
            response,
            Nsset(
                nsset_id=NssetId("kryten"),
                nsset_history_id=nsset_history_id,
                nsset_handle="HANDLE",
                dns_hosts=[DnsHost(fqdn="dns.example.com", ip_addresses=[IPv4Address("127.0.0.1")])],
            ),
        )

        request = NssetInfoRequest()
        request.nsset_id.uuid.value = "kryten"
        if nsset_history_id:
            request.nsset_history_id.uuid.value = nsset_history_id
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.Registry.Api.Nsset.Nsset/get_nsset_info",
            timeout=None,
        )

    async def test_get_nsset_info(self):
        await self._test_get_nsset_info()

    async def test_get_nsset_info_with_history_id(self):
        await self._test_get_nsset_info(nsset_history_id=NssetHistoryId("nova-5"))

    async def _test_check_dns_host(self, fqdn: str, request_fqdn: str) -> None:
        reply = CheckDnsHostReply()
        reply.data.dns_host_exists = True
        self.client.mock.return_value = make_awaitable(reply)

        self.assertTrue(await self.client.check_dns_host(fqdn))

        request = CheckDnsHostRequest()
        request.fqdn = request_fqdn
        self.assertEqual(
            self.client.mock.mock_calls,
            [call(request, method="/Fred.Registry.Api.Nsset.Nsset/check_dns_host", timeout=None)],
        )

    async def test_check_dns_host(self):
        await self._test_check_dns_host("example.org", "example.org")

    async def test_check_dns_host_idn(self):
        await self._test_check_dns_host("háčkyčárky.cz", "xn--hkyrky-ptac70bc.cz")
